/* Author: Samuel Hornus <samuel.hornus@inria.fr>
 * Copyright © Inria, 2015
 * Licence: Creative Commons CC BY-ND 3.0 license available online at
 * http://creativecommons.org/licenses/by-nd/3.0/
 */

#ifndef RAY_PHERICAL_H
#define RAY_PHERICAL_H

#include "vec.h"
#include <vector>
#include <iostream>

struct SphericalPolygonElement {
  Vec3f vertex_;
  Vec3f north_;
  void set(const Vec3f & v, const Vec3f & n) {
    vertex_ = v;
    north_ = n;
  }
  SphericalPolygonElement(){}
  SphericalPolygonElement(const Vec3f & sil)
    : north_(sil.normalized()) {}
  SphericalPolygonElement(const Vec3f & v, const Vec3f & n)
    : vertex_(v), north_(n) {}
};

struct SphericalPolygon : public std::vector<SphericalPolygonElement> {
  typedef std::vector<SphericalPolygonElement> Base;
  typedef Base::iterator iterator;
  SphericalPolygon() {
    reserve(16);
  }

  Vec3f averageDirection() const {
    // PRECONDITION : all northes are normalized.
    switch( size() ) {
      case 0 : return Vec3f(0.0f, 0.0f, 0.0f); break;
      case 1 : return begin()->north_; break;
      case 2 : return (*this)[0].north_ + (*this)[1].north_; break;
      default : {
                  Vec3f avg;
                  for( const SphericalPolygonElement & v : *this )
                    avg = avg + v.vertex_;
                  return avg;
                } break;
    }
  }

  void set_to_triangle(const Vec3f pts[3]) {
    clear();
    emplace_back(Vec3f::cross(pts[0], pts[1]).LInfNormalized(), pts[1].LInfNormalized());
    emplace_back(Vec3f::cross(pts[1], pts[2]).LInfNormalized(), pts[2].LInfNormalized());
    emplace_back(Vec3f::cross(pts[2], pts[0]).LInfNormalized(), pts[0].LInfNormalized());
  }

  void clip(const Vec3f & silVertex, SphericalPolygon & tempSpace, bool doClean = true) {
    // PRECONDITION : clipNorth, and all northes are normalized.
#define _ray_spherical_eps 1e-6f
    const int n = size();
    switch( n ) {
      case 0 : break;
      case 1 : {
                 Vec3f clipNorth = silVertex.normalized();
                 float dot = begin()->north_ | clipNorth;
                 if( dot < -0.99984769515 ) { // about one degree
                   // intersection of two almost opposite hemispheres ==> empty
                   clear();
                   break;
                 } else if( dot > 0.99984769515 ) {
                   break;
                 }
                 Vec3f v(Vec3f::cross(clipNorth, begin()->north_).LInfNormalized());
                 begin()->vertex_ = v;
                 emplace_back(v.negated(), clipNorth);
                 break;
               }
      case 2 : {
                 Vec3f clipNorth = silVertex.LInfNormalized();
                 iterator next = begin();
                 iterator cur = next++;
                 float vDot = begin()->vertex_ | clipNorth;
                 if( vDot >= _ray_spherical_eps ) {
                   // we'll get a triangle
                   next->vertex_ = Vec3f::cross(clipNorth, next->north_).LInfNormalized();
                   Vec3f v(Vec3f::cross(cur->north_, clipNorth).LInfNormalized());
                   emplace(next, v, clipNorth);
                 } else if( vDot <= - _ray_spherical_eps ) {
                   // we'll get a triangle
                   cur->vertex_ = Vec3f::cross(clipNorth, cur->north_).LInfNormalized();
                   Vec3f v(Vec3f::cross(next->north_, clipNorth).LInfNormalized());
                   emplace_back(v, clipNorth);
                 } else {
                   // we keep a moon crescent
                   float curTest(clipNorth | Vec3f::cross(cur->north_, cur->vertex_));
                   Vec3f nextTest(Vec3f::cross(next->north_, next->vertex_));
                   if( curTest > 0.0f ) {
                     if( (clipNorth | nextTest) <= 0.0f ) {
                       next->north_ = clipNorth;
                       cur->vertex_ = Vec3f::cross(next->north_, cur->north_);
                       cur->vertex_.LInfNormalize();
                       next->vertex_ = cur->vertex_;
                       next->vertex_.negate();
                     } else {
                       // the crescent is unchanged
                       //std::cerr << "kept a crescent\n";
                     }
                   } else {
                     if( (clipNorth | nextTest) > 0.0f ) {
                       cur->north_ = clipNorth;
                       next->vertex_ = Vec3f::cross(cur->north_, next->north_);
                       next->vertex_.LInfNormalize();
                       cur->vertex_ = next->vertex_;
                       cur->vertex_.negate();
                     } else {
                       //std::cerr << "killed a crescent\n";
                       clear();
                     }
                   }
                 }
                 break;
               }
      default : { // n >= 3
                  tempSpace.clear();
                  int nbKept(0);
                  iterator cur = begin();
                  Vec3f clipNorth = silVertex.LInfNormalized();
                  float nextDot, curDot = clipNorth | cur->vertex_;
                  while( cur != end() ) {
                    if( cur+1 == end() )
                      nextDot = clipNorth | begin()->vertex_;
                    else
                      nextDot = clipNorth | (cur+1)->vertex_;
                    if( curDot >= _ray_spherical_eps ) { // cur is "IN"
                      ++nbKept;
                      tempSpace.push_back(*cur);
                      if( nextDot <= -_ray_spherical_eps ) { // next is "OUT"
                        tempSpace.emplace_back(Vec3f::cross(cur->north_, clipNorth).LInfNormalized(), clipNorth);
                      }
                    } else if( curDot > -_ray_spherical_eps ) { // cur is "ON" the clipping plane
                      ++nbKept;
                      if ( nextDot <= -_ray_spherical_eps ) // next is "OUT"
                        tempSpace.emplace_back(cur->vertex_, clipNorth);
                      else
                        tempSpace.push_back(*cur);
                    } else { // cur is "OUT"
                      if ( nextDot >= _ray_spherical_eps ) { // next is "IN"
                        tempSpace.emplace_back(Vec3f::cross(clipNorth, cur->north_).LInfNormalized(), cur->north_);
                      }
                    }
                    curDot = nextDot;
                    ++cur;
                  }
                  swap(tempSpace);
                  if( (size() < 3/*too small*/) || (doClean && (nbKept == n)/*no change*/) ) {
                    clear();
                  }
                  break;
                }
    }
  }
};

#endif // RAY_PHERICAL_H
